/*!
 * asserted
 * https://github.com/alexmingoia/asserted
 */

'use strict';

/**
 * A simple extensible interface for validation assertions shipped with the
 * minimum assertions necessary for robust validations.
 *
 * @module asserted
 * @alias Assert
 */

module.exports = Assert;

/**
 * Create a new assertion with given `value`.
 *
 * Check `Assert#satisfied` and if false a failure message will be available
 * via `Assert#message`.
 *
 * @param {Mixed} value
 * @param {String=} message override default message. %v is replaced with value
 * @return {Assert}
 * @alias module:asserted
 * @constructor
 */

function Assert (value, message) {
  if (!(this instanceof Assert)) {
    return new Assert(value, message);
  }

  this.value = value;
  this.satisfied = this.assert(value);
  this.message = this.format(message || this.message, value);
};

/**
 * Throw error if assert is used directly.
 *
 * @private
 */

Assert.prototype.assert = function () {
  throw new Error(
    "Assert is not meant to be used but extended with Assert.extend()"
  );
};

/**
 * Replace %v in `message` with `value`.
 *
 * @param {String} message
 * @param {Mixed} value
 * @return {String}
 * @private
 */

Assert.prototype.format = function (message, value) {
  var replacement = value.toString ? value.toString() : value;
  return message.replace(/%v/g, replacement);
};

/**
 * Extend Assert with given `message` and assertion `fn`.
 *
 * @param {String} message %v is replaced with asserted value
 * @param {Function(value)} fn assertion function returns boolean
 * @return {Assert}
 */

Assert.extend = function (message, fn) {
  var parent = this;

  var child = function assertion (value, message) {
    if (!(this instanceof assertion)) {
      return new assertion(value, message);
    }

    return parent.apply(this, arguments);
  };

  child.extend = parent.extend;

  var Surrogate = function() {
    this.constructor = child;
  };
  Surrogate.prototype = parent.prototype;
  child.prototype = new Surrogate();

  child.prototype.assert = fn;
  child.prototype.message = message;

  child.__super__ = parent.prototype;

  return child;
};

/**
 * Create assertion that value is type of given `type`.
 *
 * @param {String} type
 * @return {Assert}
 */

Assert.TypeOf = function (type) {
  if (typeof type !== 'string') {
    throw new Error("type must be a string");
  }

  var message = "`%v` is not of type " + type;

  return this.extend(message, function (value) {
    return typeof value === type;
  });
};

/**
 * Create assertion that value is instance of given `constructor`.
 *
 * @param {Function} constructor
 * @return {Assert}
 */

Assert.InstanceOf = function (constructor) {
  if (typeof constructor !== 'function') {
    throw new Error("constructor must be a function");
  }

  var message = "`%v` is not an instance of " + constructor.name;

  return this.extend(message, function (value) {
    return (value instanceof constructor);
  });
};

/**
 * Create assertion that value is equal to given `value`.
 *
 * @param {Mixed} value
 * @return {Assert}
 */

Assert.Equals = function (value) {
  var str = value.toString ? value.toString() : value;
  var message = "`%v` is not equal to `" + str + "`";

  return this.extend(message, function (subject) {
    return subject === value;
  });
};

/**
 * Create assertion that value is in given `set`.
 *
 * @param {Array} set
 * @return {Assert}
 */

Assert.Enum = function (set) {
  if (!(set instanceof Array) || !set.length) {
    throw new Error("enum set must be a non-empty array");
  }

  var message = "`%v` is not one of " + JSON.stringify(set);

  return this.extend(message, function (value) {
    return !!~set.indexOf(value);
  });
};

/**
 * Create assertion that value array members satisfy the given `assert`.
 *
 * @param {Assert} assert
 * @return {Assert}
 */

Assert.Set = function (assert) {
  if (typeof assert !== 'function' || !assert.prototype.assert) {
    throw new Error("set assertion must be an Assert");
  }

  return this.extend(assert.prototype.message, function (value) {
    if (!(value instanceof Array)) {
      return false;
    }

    for (var i=0, l=value.length; i<l; i++) {
      var assertion = assert(value[i]);

      if (!assertion.satisfied) {
        return false;
      }
    }

    return true;
  });
};

/**
 * Create assertion that number value is within given `range`.
 *
 * @param {Object} range
 * @param {Number} range.min
 * @param {Number} range.max
 * @return {Assert}
 */

Assert.Range = function (range) {
  if (!range || !range.min || !range.max) {
    throw new Error("range must be an object with min and max");
  }

  return this.extend(
    "`%v` is not a number between " + range.min + " and " + range.max,
    function (value) {
      return typeof value === 'number' && value > range.min && value < range.max;
    });
};

/**
 * Create assertion that string or array value has given `length`.
 *
 * @param {Object|Number} length
 * @param {Number=} length.min
 * @param {Number=} length.max
 * @return {Assert}
 */

Assert.Length = function (length) {
  var message = "`%v` must have a ";

  if (typeof length === 'number') {
    message += "length of " + length;
    length = { min: length, max: length };
  } else {
    if (length.min && length.max) {
      message += "minimum length of " + length.min + " and maximum length of "
               + length.max;
    } else {
      if (length.min) {
        message += "minimum length of " + length.min;
      }
      if (length.max) {
        message += "maximum length of " + length.max;
      }
    }
  }


  return this.extend(message, function (value) {
    if (typeof value !== 'string' && !(value instanceof Array)) {
      return false;
    }

    if (length.min && value.length < length.min) {
      return false;
    }

    if (length.max && value.length > length.max) {
      return false;
    }

    return true;
  });
};

/**
 * Create assertion that value matches given `regex`.
 *
 * @param {RegExp} regex
 * @return {Assert}
 */

Assert.RegExp = function (regex) {
  var message = "`%v` does not match " + regex.toString();

  return this.extend(message, function (value) {
    return regex.test(value);
  });
};
